#include "t3f/t3f.h"
#include "t3f/memory.h"
#include "t3f/music.h"
#include "t3f/sound.h"
#include "t3f/resource.h"
#include "t3net/t3net.h"
#include "instance.h"
#include "modules/data.h"
#include "modules/dictionary.h"
#include "modules/shape.h"
#include "modules/effects.h"
#include "modules/music_queue.h"
#include "modules/video_mode.h"
#include "modules/text_entry.h"
#include "modules/leaderboards.h"
#include "modules/input.h"
#include "ui/title.h"
#include "ui/menu_leaderboards.h"
#include "ui/privacy.h"
#include "gameplay/game.h"
#include "gameplay/game_over.h"
#include "gameplay/game_intro.h"
#include "gameplay/game_music.h"
#include "gameplay/game_tally.h"
#include "gameplay/game_tutorial.h"
#include "init.h"
#include "exit.h"

static int old_display_width = 0;
static int old_display_height = 0;

/* we have to monitor the display size to check for changes since we don't
   get the correct size immediately after a resize event under Linux */
static void work_around_linux_resize_bug(void)
{
	if(al_get_display_width(t3f_display) != old_display_width || al_get_display_height(t3f_display) != old_display_height)
	{
		t3f_adjust_view(t3f_default_view, 0, 0, al_get_display_width(t3f_display), al_get_display_height(t3f_display), WL_VIRTUAL_DISPLAY_WIDTH / 2, WL_VIRTUAL_DISPLAY_HEIGHT / 2, t3f_flags);
	}
	old_display_width = al_get_display_width(t3f_display);
	old_display_height = al_get_display_height(t3f_display);
}

void wl_state_logic(int state, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	work_around_linux_resize_bug();
	switch(state)
	{
		case WL_STATE_TITLE:
		case WL_STATE_LEADERBOARDS:
		{
			wl_title_logic(data);
			break;
		}
		case WL_STATE_TITLE_CREDITS:
		{
			wl_title_credits_logic(data);
			break;
		}
		case WL_STATE_TITLE_EXIT:
		{
			wl_title_exit_logic(data);
			break;
		}
		case WL_STATE_TO_GAME:
		{
			wl_title_to_game_logic(data);
			break;
		}
		case WL_STATE_GAME_INTRO:
		{
			wl_game_intro_logic(data);
			if(instance->cinema && instance->cinema->break_points)
			{
				wl_handle_cinema_skip(instance->cinema, instance);
			}
			break;
		}
		case WL_STATE_GAME:
		{
			wl_game_logic(data);
			break;
		}
		case WL_STATE_MUSIC:
		{
			wl_game_music_logic(data);
			break;
		}
		case WL_STATE_GAME_TUTORIAL:
		{
			wl_game_tutorial_logic(data);
			break;
		}
		case WL_STATE_GAME_OVER:
		{
			wl_game_over_logic(data);
			break;
		}
		case WL_STATE_GAME_PAUSED:
		{
			wl_game_paused_logic(data);
			if(instance->cinema && instance->cinema->break_points)
			{
				wl_handle_cinema_skip(instance->cinema, instance);
			}
			break;
		}
		case WL_STATE_GLOBAL_PAUSE:
		{
			wl_title_process_menu(WL_MENU_GAME_EXIT, data);
			break;
		}
		case WL_STATE_GAME_CONTINUE:
		{
			wl_game_continue_logic(data);
			if(instance->cinema && instance->cinema->break_points)
			{
				wl_handle_cinema_skip(instance->cinema, instance);
			}
			break;
		}
		case WL_STATE_GAME_QUIT:
		{
			wl_game_quit_logic(data);
			break;
		}
		case WL_STATE_MUSIC_QUIT:
		{
			wl_game_music_quit_logic(data);
			break;
		}
		case WL_STATE_GAME_WIN:
		{
			wl_game_win_logic(data);
			if(instance->cinema && instance->cinema->break_points)
			{
				wl_handle_cinema_skip(instance->cinema, instance);
			}
			break;
		}
		case WL_STATE_DEMO_WIN:
		{
			wl_demo_win_logic(data);
			break;
		}
		case WL_STATE_GAME_OVER_WAIT:
		{
			wl_game_over_wait_logic(data);
			break;
		}
		case WL_STATE_GAME_TALLY:
		{
			wl_game_tally_logic(data);
			break;
		}
		case WL_STATE_GAME_DONE:
		{
			wl_game_done_logic(data);
			break;
		}
		case WL_STATE_PRIVACY:
		{
			wl_privacy_logic(data);
			break;
		}
		default:
		{
			break;
		}
	}
}

void wl_logic(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	char buf[256];
	bool viewing_leaderboard = false;

	t3f_steam_integration_logic();
	if(instance->exit_tick <= 0)
	{
		if(instance->input_device_lost)
		{
			wl_select_input(wl_get_default_input_type(), 0, data);
			instance->input_device_lost = false;
		}
		t3f_update_input_handler_state(instance->input_handler);
		wl_read_user_input(data);

		/* save newly fetched user key */
		if(instance->new_user_key)
		{
			al_set_config_value(t3f_user_data, "Leaderboards", "user_key", instance->user_key);
			t3f_save_user_data();
			instance->new_user_key = false;
		}

		/* start leaderboard load if download wanted */
		if(instance->leaderboard_want_download)
		{
			if(!t3f_object_loading(instance->leaderboard_loader))
			{
				/* free previously loaded leaderboard */
				if(instance->leaderboard)
				{
					t3net_destroy_leaderboard(instance->leaderboard);
				}
				instance->leaderboard = t3f_load_object(instance->leaderboard_loader, _wl_leaderboard_download_proc, _wl_leaderboard_destroy_proc, data, true);
				instance->leaderboard_want_download = false;
			}
			else if(t3f_object_ready(instance->leaderboard_loader))
			{
				instance->leaderboard = t3f_fetch_object(instance->leaderboard_loader);
			}
		}
		/* fetch leaderboard data loaded from thread */
		else if(t3f_object_ready(instance->leaderboard_loader))
		{
			instance->leaderboard = t3f_fetch_object(instance->leaderboard_loader);
			if(instance->state == WL_STATE_LEADERBOARDS)
			{
				viewing_leaderboard = true;
			}
			else if(instance->state == WL_STATE_GAME_DONE && instance->current_menu == WL_MENU_DYNAMIC)
			{
				viewing_leaderboard = true;
			}
			else if(instance->state == WL_STATE_GAME_OVER_WAIT && instance->current_menu == WL_MENU_DYNAMIC)
			{
				viewing_leaderboard = true;
			}
			if(instance->leaderboard && viewing_leaderboard)
			{
				wl_create_leaderboard_menu(instance->leaderboard_mode, instance->leaderboard_type, instance);
			}
			else
			{
				t3net_destroy_leaderboard(instance->leaderboard);
				instance->leaderboard = NULL;
			}
		}

		if(t3f_key_pressed(ALLEGRO_KEY_TILDE) || t3f_key_pressed(104))
		{
			sprintf(buf, "screenshot_%d.png", instance->tick);
			al_save_bitmap(buf, al_get_backbuffer(t3f_display));
			t3f_use_key_press(ALLEGRO_KEY_TILDE);
			t3f_use_key_press(104);
		}
		wl_process_music_queue(instance->music_queue);
		wl_state_logic(instance->state, data);
		wl_process_effects(instance->effects_data);
		wl_message_logic(&instance->award_messages, false);
	}

	if(instance->exit_tick > 0)
	{
		wl_process_music_queue(instance->music_queue);
		instance->exit_tick--;
		if(instance->exit_tick <= 0)
		{
			t3f_exit();
		}
	}

	instance->tick++;
}

void wl_state_render(int state, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	switch(state)
	{
		case WL_STATE_TITLE:
		case WL_STATE_LEADERBOARDS:
		{
			wl_title_render(data);
			break;
		}
		case WL_STATE_TITLE_CREDITS:
		{
			wl_title_credits_render(data);
			break;
		}
		case WL_STATE_TITLE_EXIT:
		{
			wl_title_exit_render(data);
			break;
		}
		case WL_STATE_TO_GAME:
		{
			wl_title_to_game_render(data);
			break;
		}
		case WL_STATE_GAME_INTRO:
		{
			wl_game_intro_render(data);
			break;
		}
		case WL_STATE_GAME:
		{
			wl_game_render(data);
			break;
		}
		case WL_STATE_MUSIC:
		{
			wl_game_music_render(data);
			break;
		}
		case WL_STATE_GAME_TUTORIAL:
		{
			wl_game_tutorial_render(data);
			break;
		}
		case WL_STATE_GAME_OVER:
		{
			wl_game_over_render(data);
			break;
		}
		case WL_STATE_GAME_PAUSED:
		{
			wl_game_paused_render(data);
			break;
		}
		case WL_STATE_GLOBAL_PAUSE:
		{
			wl_state_render(instance->global_pause_state, data);
			al_draw_filled_rectangle(0, 0, t3f_virtual_display_width, t3f_virtual_display_height, al_map_rgba_f(0.0, 0.0, 0.0, 0.5));
			if(instance->exit_tick < 0)
			{
				t3f_render_gui(instance->menu[WL_MENU_GAME_EXIT], T3F_GUI_USER_FLAG);
				t3f_render_gui(instance->menu[WL_MENU_GAME_EXIT], 0);
			}
			break;
		}
		case WL_STATE_GAME_CONTINUE:
		{
			wl_game_intro_render(data);
			break;
		}
		case WL_STATE_GAME_QUIT:
		{
			wl_game_intro_render(data);
			break;
		}
		case WL_STATE_MUSIC_QUIT:
		{
			wl_game_music_render(data);
			break;
		}
		case WL_STATE_GAME_WIN:
		{
			wl_game_win_render(data);
			break;
		}
		case WL_STATE_DEMO_WIN:
		{
			wl_demo_win_render(data);
			break;
		}
		case WL_STATE_GAME_OVER_WAIT:
		{
			wl_game_over_wait_render(data);
			break;
		}
		case WL_STATE_GAME_TALLY:
		{
			wl_game_tally_render(data);
			break;
		}
		case WL_STATE_GAME_DONE:
		{
			wl_game_done_render(data);
			break;
		}
		case WL_STATE_PRIVACY:
		{
			wl_privacy_render(data);
			break;
		}
		default:
		{
			break;
		}
	}
}

void wl_render(void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	float a;

	/* when exiting, render exit state instead of current state */
	wl_state_render(instance->state, data);
	wl_render_effects(instance->effects_data);
	al_hold_bitmap_drawing(true);
	wl_render_messages(&instance->award_messages, false, data);
	al_hold_bitmap_drawing(false);

	if(instance->exit_tick >= 0)
	{
		a = 1.0 - (float)instance->exit_tick / (float)WL_FADE_TICKS;
		al_draw_filled_rectangle(0, 0, t3f_virtual_display_width, t3f_virtual_display_height, al_map_rgba_f(0.0, 0.0, 0.0, a));
	}
}

int main(int argc, char * argv[])
{
	WL_INSTANCE * instance;

	instance = wl_create_instance();
	if(!instance)
	{
		printf("Failed to create instance!\n");
		goto fail;
	}

	if(!wl_initialize(argc, argv, instance))
	{
		printf("Failed to initialize game!\n");
		goto fail;
	}
//	_wl_fix_animations();
	t3f_run();
	wl_exit(instance);
	t3f_finish();
	wl_destroy_instance(instance);
	return 0;

	fail:
	{
		return -1;
	}
}
