#include "t3f/t3f.h"
#include "t3f/music.h"
#include "t3f/resource.h"
#include "instance.h"
#include "level.h"
#include "music_queue.h"

int wl_get_level_slot(int level)
{
	return level & 1;
}

/* load level shapes for the specified level */
bool wl_load_level_shapes(WL_LEVEL * lp, ALLEGRO_CONFIG * scp, ALLEGRO_PATH * path, int flags)
{
	const char * val;
	int i;
	char buf[256] = {0};
	bool loaded_shapes = false;
	float source_width, source_height;
	float target_width, target_height;
	float scale_x, scale_y;
	
	val = al_get_config_value(scp, "Level Data", "bitmap");
	if(val)
	{
		al_set_path_filename(path, val);
		lp->shape_image = t3f_load_bitmap(al_path_cstr(path, '/'), 0, flags & WL_LEVEL_LOAD_FLAG_THREADED);
		if(lp->shape_image)
		{
			/* get target size so we can scale our shapes to match the loaded bitmap */
			source_width = lp->shape_image->target_width;
			source_height = lp->shape_image->target_height;
			target_width = source_width;
			target_height = source_height;
			val = al_get_config_value(scp, "Level Data", "target_width");
			if(val)
			{
				target_width = atof(val);
			}
			val = al_get_config_value(scp, "Level Data", "target_height");
			if(val)
			{
				target_height = atof(val);
			}
			scale_x = source_width / target_width;
			scale_y = source_height / target_height;
			
			val = al_get_config_value(scp, "Level Data", "shapes");
			if(val)
			{
				lp->shapes = atoi(val);
				for(i = 0; i < lp->shapes; i++)
				{
					sprintf(buf, "shape %d", i);
					val = al_get_config_value(scp, buf, "x");
					if(val)
					{
						lp->shape[i].x = atof(val) * scale_x;
					}
					val = al_get_config_value(scp, buf, "y");
					if(val)
					{
						lp->shape[i].y = atof(val) * scale_y;
					}
					val = al_get_config_value(scp, buf, "w");
					if(val)
					{
						lp->shape[i].target_width = atof(val);
						lp->shape[i].width = lp->shape[i].target_width * scale_x;
					}
					val = al_get_config_value(scp, buf, "h");
					if(val)
					{
						lp->shape[i].target_height = atof(val);
						lp->shape[i].height = lp->shape[i].target_height * scale_y;
					}
				}
				loaded_shapes = true;
				lp->shapes_used = false;
			}
		}
	}
	return loaded_shapes;
}

static char * _wl_get_music_path_by_level_name(const char * name, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	ALLEGRO_PATH * path;
	ALLEGRO_CONFIG * config;
	const char * val;
	int i;
	char * ret = NULL;

	for(i = 0; i < instance->level_list->entry_count; i++)
	{
		if(!strcmp(instance->level_list->entry[i].name, name))
		{
			path = al_create_path(instance->level_list->entry[i].path);
			if(path)
			{
				config = al_load_config_file(al_path_cstr(path, '/'));
				if(config)
				{
					val = al_get_config_value(config, "Level Data", "music");
					if(val)
					{
						al_set_path_filename(path, val);
						ret = strdup(al_path_cstr(path, '/'));
					}
					al_destroy_config(config);
				}
				al_destroy_path(path);
			}
		}
	}
	return ret;
}

void wl_play_level_music(WL_LEVEL * lp, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;

	if(lp->music_path)
	{
		wl_queue_music(instance->music_queue, lp->music_path);
		wl_finish_current_song(instance->music_queue, 0.0);
	}
}

static char * _wl_get_music_path(ALLEGRO_PATH * path, const char * fn, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	bool music_found = false;

	/* play music by level name if ':' prefix used */
	if(fn[0] == ':')
	{
		return _wl_get_music_path_by_level_name(&fn[1], data);
	}

	/* play music by filename in path relative to level INI file */
	else
	{
		al_set_path_filename(path, fn);
		if(al_filename_exists(al_path_cstr(path, '/')))
		{
			return strdup(al_path_cstr(path, '/'));
		}
	}

	/* play default music if music not found or specified */
	if(!music_found)
	{
		return _wl_get_music_path_by_level_name(instance->level_list->entry[0].name, data);
	}

	return NULL;
}

WL_LEVEL * wl_load_level(int level, int flags, void * data)
{
	WL_INSTANCE * instance = (WL_INSTANCE *)data;
	ALLEGRO_CONFIG * cp = NULL;
	ALLEGRO_PATH * path = NULL;
	const char * val;
	float r = 0.0, g = 0.0, b = 0.0, a = 1.0;
	char buf[1024];
	WL_LEVEL * lp;
	
	#ifdef ALLEGRO_ANDROID
		flags &= ~WL_LEVEL_LOAD_FLAG_THREADED;
		al_stop_timer(t3f_timer);
	#endif

	lp = malloc(sizeof(WL_LEVEL));
	if(!lp)
	{
		goto fail;
	}
	memset(lp, 0, sizeof(WL_LEVEL));

	path = al_create_path(instance->level_list->entry[level].path);
	if(!path)
	{
		goto fail;
	}
	cp = al_load_config_file(instance->level_list->entry[level].path);
	if(cp)
	{
		val = al_get_config_value(cp, "Level Data", "music");
		if(val)
		{
			lp->music_path = _wl_get_music_path(path, val, data);
		}
		
		/* load everything else unless we just want the music */
		if(!(flags & WL_LEVEL_LOAD_FLAG_MUSIC_ONLY))
		{
			val = al_get_config_value(cp, "Level Data", "bg");
			if(val)
			{
				al_set_path_filename(path, val);
				sprintf(buf, "%s#Size %1.1f", al_path_cstr(path, '/'), instance->graphics_scale);
				lp->bg_animation = t3f_load_animation(buf, 0, flags & WL_LEVEL_LOAD_FLAG_THREADED);
				if(!lp->bg_animation)
				{
					goto fail;
				}
			}
			else
			{
				goto fail;
			}
			val = al_get_config_value(cp, "Level Data", "name");
			if(val)
			{
				lp->name = strdup(val);
			}
			lp->number = level;
			val = al_get_config_value(cp, "Level Data", "color_r");
			if(val)
			{
				r = atof(val);
			}
			val = al_get_config_value(cp, "Level Data", "color_g");
			if(val)
			{
				g = atof(val);
			}
			val = al_get_config_value(cp, "Level Data", "color_b");
			if(val)
			{
				b = atof(val);
			}
			val = al_get_config_value(cp, "Level Data", "color_a");
			if(val)
			{
				a = atof(val);
			}
			lp->color = al_map_rgba_f(r, g, b, a);
			
			lp->z = 0.0;
			lp->vz = 0.0;
			val = al_get_config_value(cp, "Level Data", "zoom");
			if(val)
			{
				lp->vz = atof(val);
			}
			
			lp->a = 0.0;
			lp->va = 0.0;
			val = al_get_config_value(cp, "Level Data", "rotate");
			if(val)
			{
				lp->va = atof(val);
			}
			
			if(flags & WL_LEVEL_LOAD_FLAG_THREADED)
			{
				lp->cx = -10000;
			}
			else
			{
				lp->cx = lp->bg_animation->data->frame[0]->width / 2;
			}
			val = al_get_config_value(cp, "Level Data", "cx");
			if(val)
			{
				lp->cx = atof(val);
			}
			if(flags & WL_LEVEL_LOAD_FLAG_THREADED)
			{
				lp->cy = -10000;
			}
			else
			{
				lp->cy = lp->bg_animation->data->frame[0]->height / 2;
			}
			val = al_get_config_value(cp, "Level Data", "cy");
			if(val)
			{
				lp->cy = atof(val);
			}
			lp->ox = WL_VIRTUAL_DISPLAY_WIDTH / 2;
			val = al_get_config_value(cp, "Level Data", "ox");
			if(val)
			{
				lp->ox = atof(val);
			}
			lp->oy = 240.0;
			val = al_get_config_value(cp, "Level Data", "oy");
			if(val)
			{
				lp->oy = atof(val);
			}
			
			/* load music info */
			val = al_get_config_value(cp, "Level Data", "artist");
			if(val)
			{
				lp->artist = strdup(val);
			}
			else
			{
				lp->artist = strdup("");
			}
			val = al_get_config_value(cp, "Level Data", "music artist");
			if(val)
			{
				lp->music_artist = strdup(val);
			}
			else
			{
				lp->music_artist = strdup("");
			}

			val = al_get_config_value(cp, "Level Data", "music title");
			if(val)
			{
				lp->music_title = strdup(val);
			}
			else
			{
				lp->music_title = strdup("");
			}
			
			/* load shapes */
			val = al_get_config_value(cp, "Level Data", "shapes");
			if(val)
			{
				if(!wl_load_level_shapes(lp, cp, path, flags))
				{
					printf("no shapes\n");
					goto fail;
				}
			}
		}
		al_destroy_config(cp);
	}
	#ifdef ALLEGRO_ANDROID
		al_start_timer(t3f_timer);
	#endif
	return lp;

	fail:
	{
		#ifdef ALLEGRO_ANDROID
			al_start_timer(t3f_timer);
		#endif
		if(cp)
		{
			al_destroy_config(cp);
		}
		if(lp)
		{
			wl_destroy_level(lp);
		}
		if(path)
		{
			al_destroy_path(path);
		}
		return NULL;
	}
}

void wl_destroy_level(WL_LEVEL * lp)
{
	if(lp)
	{
		if(lp->music_title)
		{
			free(lp->music_title);
		}
		if(lp->music_artist)
		{
			free(lp->music_artist);
		}
		if(lp->artist)
		{
			free(lp->artist);
		}
		if(lp->bg_animation)
		{
			t3f_destroy_animation(lp->bg_animation);
			lp->bg_animation = NULL;
		}
		if(lp->shape_image)
		{
			t3f_destroy_bitmap(lp->shape_image);
			lp->shape_image = NULL;
		}
		if(lp->name)
		{
			free(lp->name);
		}
		if(lp->music_path)
		{
			free(lp->music_path);
		}
		free(lp);
	}
}

void wl_update_level(WL_LEVEL * lp)
{
	if(lp->bg_animation)
	{
		if(t3f_object_ready(lp->bg_animation->object_loader))
		{
			t3f_update_animation(lp->bg_animation);
			if(lp->cx <= -10000.0)
			{
				lp->cx = lp->bg_animation->data->frame[0]->width / 2;
			}
			if(lp->cy <= -10000.0)
			{
				lp->cy = lp->bg_animation->data->frame[0]->height / 2;
			}
		}
	}
	if(lp->shape_image)
	{
		t3f_update_bitmap(lp->shape_image);
	}
}
