#include "t3f/music.h"
#include "instance.h"
#include "music_queue.h"
#include "data.h"

WL_MUSIC_QUEUE * wl_create_music_queue(void)
{
	WL_MUSIC_QUEUE * qp;

	qp = malloc(sizeof(WL_MUSIC_QUEUE));
	if(!qp)
	{
		goto fail;
	}
	memset(qp, 0, sizeof(WL_MUSIC_QUEUE));
  qp->music_queue_state = WL_MUSIC_QUEUE_STATE_OFF;
  qp->music_fade_level = 1.0;
  qp->music_fade_speed = 1.0 / WL_LEVEL_FADE_FRAMES;
	return qp;

	fail:
	{
		return NULL;
	}
}

void wl_destroy_music_queue(WL_MUSIC_QUEUE * qp)
{
	free(qp);
}

static bool wl_delete_music_from_queue(WL_MUSIC_QUEUE * qp, int index)
{
	int i;
	
	if(index < qp->music_queue_size)
	{
		/* move all music past index to previous slot */
		for(i = index; i < qp->music_queue_size - 1; i++)
		{
			strncpy(qp->music_queue_fn[i], qp->music_queue_fn[i + 1], 256);
		}
		qp->music_queue_size--; // decrease queue size
		return true;
	}
	return false;
}

/* play next song in the queue */
static bool wl_play_queued_music(WL_MUSIC_QUEUE * qp)
{
	if(qp->music_queue_size > 0)
	{
		t3f_set_new_music_volume(qp->base_volume);
		t3f_play_music(qp->music_queue_fn[0]);
		wl_delete_music_from_queue(qp, 0);
		return true;
	}
	return false;
}

/* clear queue */
void wl_clear_music_queue(WL_MUSIC_QUEUE * qp)
{
	qp->music_queue_size = 0;
	qp->music_queue_state = WL_MUSIC_QUEUE_STATE_OFF;
	t3f_stop_music();
}

/* interrupt the queue with a new song, use for everything but in-game */
bool wl_play_music(WL_MUSIC_QUEUE * qp, const char * fn)
{
	bool ret;

	wl_clear_music_queue(qp);
	t3f_set_music_volume(qp->base_volume);

	ret = t3f_play_music(fn);
	if(ret)
	{
		qp->music_queue_state = WL_MUSIC_QUEUE_STATE_PLAYING;
	}
	else
	{
		qp->music_queue_state = WL_MUSIC_QUEUE_STATE_OFF;
	}
	return ret;
}

/* queue up a song */
bool wl_queue_music(WL_MUSIC_QUEUE * qp, const char * fn)
{
	if(qp->music_queue_size < WL_MUSIC_QUEUE_MAX_SIZE)
	{
		strncpy(qp->music_queue_fn[qp->music_queue_size], fn, 256);
		qp->music_queue_size++;
		return true;
	}
	return false;
}

/* begin fading current song in preparation for the next song */
void wl_finish_current_song(WL_MUSIC_QUEUE * qp, float fade_speed)
{
	if(qp->music_queue_state == WL_MUSIC_QUEUE_STATE_PLAYING)
	{
		qp->music_fade_level = 1.0;
		qp->music_queue_state = WL_MUSIC_QUEUE_STATE_TRANSITIONING;
		if(fade_speed <= 0.0)
		{
		  qp->music_fade_speed = 1.0 / WL_LEVEL_FADE_FRAMES;
		}
		else
		{
			qp->music_fade_speed = fade_speed;
		}
	}
	else if(qp->music_queue_state == WL_MUSIC_QUEUE_STATE_TRANSITIONING)
	{
		if(fade_speed <= 0)
		{
			qp->music_fade_speed = (1.0 - qp->music_fade_level) / (float)WL_LEVEL_FADE_FRAMES;
		}
		else
		{
			qp->music_fade_speed = fade_speed * (1.0 - qp->music_fade_level);
		}
	}
}

void wl_process_music_queue(WL_MUSIC_QUEUE * qp)
{
	switch(qp->music_queue_state)
	{
		/* if music is off and there is music in the queue, begin playback */
		case WL_MUSIC_QUEUE_STATE_OFF:
		{
			if(qp->music_queue_size > 0)
			{
				if(wl_play_queued_music(qp))
				{
					qp->music_queue_state = WL_MUSIC_QUEUE_STATE_PLAYING;
				}
			}
			break;
		}
		
		/* fade out current music and start next song when done */
		case WL_MUSIC_QUEUE_STATE_TRANSITIONING:
		{
			qp->music_fade_level -= qp->music_fade_speed;
			
			/* if we have finished fading, play the next song */
			if(qp->music_fade_level <= 0)
			{
				if(wl_play_queued_music(qp))
				{
					qp->music_queue_state = WL_MUSIC_QUEUE_STATE_PLAYING;
				}
				else
				{
					t3f_stop_music();
					qp->music_queue_state = WL_MUSIC_QUEUE_STATE_OFF;
				}
			}
			
			/* otherwise, set the new volume level */
			else
			{
				t3f_set_music_volume(qp->music_fade_level * qp->base_volume);
			}
			break;
		}
	}
}

int wl_get_music_queue_state(WL_MUSIC_QUEUE * qp)
{
	return qp->music_queue_state;
}

void wl_set_music_fade_speed(WL_MUSIC_QUEUE * qp, float speed)
{
	qp->music_fade_speed = speed;
}

void wl_set_music_base_volume(WL_MUSIC_QUEUE * qp, float vol)
{
	qp->base_volume = vol;
	if(qp->music_queue_state == WL_MUSIC_QUEUE_STATE_TRANSITIONING)
	{
		t3f_set_music_volume(qp->base_volume * qp->music_fade_level);
	}
	else
	{
		t3f_set_music_volume(qp->base_volume);
	}
}
